// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./Reputation.sol";
import "hardhat/console.sol";

struct Citation {
    int weightPercent;
    uint targetPostIndex;
}

struct Post {
    uint id;
    address sender;
    address author;
    string contentId;
    Citation[] citations;
    uint reputation;
}

contract Forum is Reputation {
    mapping(uint => Post) public posts;
    uint public postCount;
    mapping(uint => mapping(uint => int)) _edgeBalances;

    event PostAdded(uint postIndex);

    // Forum parameters
    // TODO: Make depth limit configurable; take as param in _onValidatePost callback
    uint depthLimit = 3;

    function addPost(
        address author,
        string calldata contentId,
        Citation[] calldata citations
    ) external returns (uint postIndex) {
        postIndex = postCount++;
        Post storage post = posts[postIndex];
        post.author = author;
        post.sender = msg.sender;
        post.id = postIndex;
        post.contentId = contentId;
        for (uint i = 0; i < citations.length; i++) {
            post.citations.push(citations[i]);
        }
        int totalCitationWeightPos;
        int totalCitationWeightNeg;
        for (uint i = 0; i < post.citations.length; i++) {
            int weight = post.citations[i].weightPercent;
            require(weight >= -100, "Each citation weight must be >= -100");
            require(weight <= 100, "Each citation weight must be <= 100");
            if (weight > 0) totalCitationWeightPos += weight;
            else totalCitationWeightNeg += weight;
        }
        require(
            totalCitationWeightPos <= 100,
            "Sum of positive citations must be <= 100"
        );
        require(
            totalCitationWeightNeg >= -100,
            "Sum of negative citations must be >= -100"
        );
        emit PostAdded(postIndex);
    }

    function _onValidatePost(uint postIndex, uint amount) internal {
        _propagateReputation(postIndex, int(amount), false, 0);
    }

    function _handleCitation(
        uint postIndex,
        Citation memory citation,
        int amount,
        bool initialNegative,
        uint depth
    ) internal returns (int outboundAmount) {
        outboundAmount = (amount * citation.weightPercent) / 100;
        if (citation.targetPostIndex == type(uint256).max) {
            // Incineration
            require(
                outboundAmount >= 0,
                "Leaching from incinerator is forbidden"
            );
            _burn(address(this), uint(outboundAmount));
            return outboundAmount;
        }
        int balanceToOutbound = _edgeBalances[postIndex][
            citation.targetPostIndex
        ];
        if (initialNegative) {
            if (outboundAmount < 0) {
                outboundAmount = outboundAmount > -balanceToOutbound
                    ? outboundAmount
                    : -balanceToOutbound;
            } else {
                outboundAmount = outboundAmount < -balanceToOutbound
                    ? outboundAmount
                    : -balanceToOutbound;
            }
        }
        int refund = _propagateReputation(
            citation.targetPostIndex,
            outboundAmount,
            initialNegative || (depth == 0 && citation.weightPercent < 0),
            depth + 1
        );
        outboundAmount -= refund;
        _edgeBalances[postIndex][citation.targetPostIndex] += outboundAmount;
    }

    function _propagateReputation(
        uint postIndex,
        int amount,
        bool initialNegative,
        uint depth
    ) internal returns (int refundToInbound) {
        if (depth >= depthLimit) {
            return amount;
        }
        Post storage post = posts[postIndex];
        // Propagate negative citations first
        for (uint i = 0; i < post.citations.length; i++) {
            if (post.citations[i].weightPercent < 0) {
                amount -= _handleCitation(
                    postIndex,
                    post.citations[i],
                    amount,
                    initialNegative,
                    depth
                );
            }
        }
        // Now propagate positive citations
        for (uint i = 0; i < post.citations.length; i++) {
            if (post.citations[i].weightPercent > 0) {
                amount -= _handleCitation(
                    postIndex,
                    post.citations[i],
                    amount,
                    initialNegative,
                    depth
                );
            }
        }
        if (amount > 0) {
            _update(address(this), post.author, uint(amount));
            post.reputation += uint(amount);
        } else {
            // Prevent reputation from being reduced below zero
            if (int(post.reputation) + amount >= 0) {
                if (balanceOf(post.author) >= uint(-amount)) {
                    _update(post.author, address(this), uint(-amount));
                } else {
                    // Author has already lost some REP gained from this post.
                    // That means other DAO members have earned it for policing.
                    // We need to refund the difference here to ensure accurate bookkeeping
                    refundToInbound = amount + int(balanceOf(post.author));
                    _update(post.author, address(this), balanceOf(post.author));
                }
                post.reputation -= uint(-amount);
            } else {
                refundToInbound = int(post.reputation) + amount;
                if (balanceOf(post.author) >= post.reputation) {
                    _update(post.author, address(this), post.reputation);
                } else {
                    // If author has already lost reputation that was gained from this post,
                    // that means other DAO members gained it through policing.
                    // We have to increase the magnitude of the amount we're "refunding", which is expressed as a negative number.
                    // This has the effect of preserving the sum of all members' REP.
                    // However, we still set the post reputation all the way to zero.
                    // So we end up decreasing the sum of all posts' REP.
                    refundToInbound -= int(
                        post.reputation - balanceOf(post.author)
                    );
                    _update(post.author, address(this), balanceOf(post.author));
                }
                post.reputation = 0;
            }
        }
    }
}
